#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include "cdecl.h"
#include "typemap.h"

/* declare [ident] as [storage] [function specs] [other crap] [qualifiers] [type specs] */

static size_t
voutput(char *buf, size_t n, size_t off, const char *fmt, va_list ap)
{
	if (off >= n)
		return vsnprintf(NULL, 0, fmt, ap);
	return vsnprintf(buf+off, n-off, fmt, ap);
}

static size_t output(char *buf, size_t n, size_t off, const char *fmt, ...)
{
	va_list ap;
	size_t ret;

	va_start(ap, fmt);
	ret = voutput(buf, n, off, fmt, ap);
	va_end(ap);

	return ret;
}

/* Renders the type qualifiers and type specifiers in canonical form. */
static size_t
explain_post_specs(char *buf, size_t n, struct cdecl_declspec *s)
{
	unsigned long qualmap = 0, typemap;
	const char *tag = NULL;
	size_t ret = 0;

	typemap = cdecl__build_typemap(s);
	if (typemap == -1)
		return 0;

	for (struct cdecl_declspec *c = s; c; c = c->next) {
		switch (cdecl_spec_kind(c)) {
		case CDECL_SPEC_QUAL:
			qualmap |= 1ul << (c->type & 0xff);
			break;
		case CDECL_SPEC_TYPE:
			/* Valid C types have at most one identifier. */
			if (c->ident)
				tag = c->ident;
			break;
		}
	}

	if (qualmap & (1ul << (CDECL_QUAL_RESTRICT & 0xff)))
		ret += output(buf, n, ret, "restrict ");
	if (qualmap & (1ul << (CDECL_QUAL_VOLATILE & 0xff)))
		ret += output(buf, n, ret, "volatile ");
	if (qualmap & (1ul << (CDECL_QUAL_CONST & 0xff)))
		ret += output(buf, n, ret, "const ");

	ret += output(buf, n, ret, "%s", cdecl__explain_typemap(typemap));
	if (tag)
		ret += output(buf, n, ret, " %s", tag);
	return ret;
}

static const char *explain_storage(unsigned spec)
{
	switch (spec) {
	case CDECL_STOR_TYPEDEF:
		return "typedef";
	case CDECL_STOR_EXTERN:
		return "extern";
	case CDECL_STOR_STATIC:
		return "static";
	case CDECL_STOR_AUTO:
		return "auto";
	case CDECL_STOR_REGISTER:
		return "register";
	default:
		abort();
	}
}

/* Renders the storage-class and function specifiers in canonical form. */
static size_t explain_pre_specs(char *buf, size_t n, struct cdecl_declspec *s)
{
	unsigned long funcmap = 0;
	const char *storage = NULL;
	size_t ret = 0;

	for (struct cdecl_declspec *c = s; c; c = c->next) {
		switch (cdecl_spec_kind(c)) {
		case CDECL_SPEC_FUNC:
			funcmap |= 1ul << (c->type & 0xff);
			break;
		case CDECL_SPEC_STOR:
			/* Valid C declarations have at most one
			 * storage-class specifier. */
			storage = explain_storage(c->type);
			break;
		}
	}

	if (storage)
		ret += output(buf, n, ret, "%s", storage);
	if (funcmap & (1ul << (CDECL_FUNC_INLINE & 0xff)))
		ret += output(buf, n, ret, "%.*s%s", !!ret, "", "inline");
	return ret;
}

static size_t advance(char **buf, size_t *n, size_t amount)
{
	if (!amount)
		return 0;

	if (amount >= *n) {
		*n   = 0;
		*buf = 0;
	} else {
		(*buf)[amount] = ' ';
		if (amount + 1 >= *n) {
			*buf = 0;
			*n = 0;
		} else {
			*buf += amount + 1;
			*n   -= amount + 1;
		}
	}

	return amount + 1;
}

size_t cdecl_explain(char *buf, size_t n, struct cdecl *decl)
{
	size_t ret = 0, rc;

	/*
	 * XXX: This won't work when we add support for more complicated
	 * declarators.
	 */
	rc = output(buf, n, 0, "declare %s as", decl->declarators->ident);
	ret += advance(&buf, &n, rc);

	rc = explain_pre_specs(buf, n, decl->specifiers);
	ret += advance(&buf, &n, rc);

	return ret + explain_post_specs(buf, n, decl->specifiers);
}
